/*
	EmissionLineIntegration.js: Narrow-band application on RGB images
================================================================================

Schmalband-Filter werden in der Astronomie eingesetzt, um die Intensität von
Emissionsstrahlern zu messen. Da diese Filter einen Durchlassbereich von einigen
Nanometer (nm) haben, addieren sich Fehler, die aus eng benachbarten Quellen
stammen. So ist z.B. Ein H-alpha Filter mit 6 nm Bandbreite noch aufnahmefähig
für die NII Linie, deren Wellenlänge nur einen Abstand von 2 nm hat.
Die Continuum Subtraction (kurz CS) soll hier helfen, die Strahlung aus dem
Continuum herauszufiltern. Das Ergebnis kann dann für die Berechnung der
Intensität eines Objektes genutzt werden. In der Astrofotografie erzeugt
die Addition des Differenzbildes auf einen Breitbandkanal eine merkliche
Verstärkung der Objekte.

Formeln

	NB 	= Schmalband
	BB 	= Breitband
	BB*	= Breitband intensiviert
	mue	= Anteil Continuum im Schmalband

	Ie 	= NB - mue x BB							(1)
	BB*	= BB + Ie									(2)


Narrow-band filters see use in astronomy to measure the intensity of emission
emitters. Since these filters have a range of bandwidth of around a few
nanometers (nm), errors from neighbouring sources will accumulate. For example,
a H-alpha-filter with 6nm bandwidth is receptive to the NII line with a
wavelength distance of 2nm. This is where the continuum subtraction (CS) helps
filter the emission out of the continuum. The resulting data can be used to
calculate the intensity of an object. In astro-photography, the addition of
the difference image onto a broadband channel generates a noticeable
amplification of the objects.

Formula

	NB 	= narrow band
	BB 	= braod band
	BB*	= braod band intensified
	mue	= fraction of continuum found in narrow band

	Ie 	= NB - mue x BB							(1)
	BB*	= BB + Ie									(2)

================================================================================


 Input

 1. Active window, RGB after ChannelCombination

 2. A narrowband at least for one channel

	It is recommended, that a DynamicBackroundExtraction-process has been
	applied to RGB and NB-channel[s] prior to ContinuumSub

 Copyright Hartmut V. Bornemann, 2017, 2018

*/

#feature-id    Utilities > EmissionLineIntegration

#include <pjsr/StdButton.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/NumericControl.jsh>
#include <pjsr/Sizer.jsh>
#include <pjsr/Slider.jsh>
#include <pjsr/UndoFlag.jsh>
#include <pjsr/SampleType.jsh>
#include <pjsr/DataType.jsh>
#include <pjsr/Color.jsh>
#include <pjsr/FileMode.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/StdCursor.jsh>
#include <pjsr/MaskMode.jsh>
#include <pjsr/ColorSpace.jsh>
#include <pjsr/FontFamily.jsh>
#include <pjsr/ImageOp.jsh>


#define TITLE "Emission Line Integration"
#define VERSION "1.1.0"
// Shadows clipping point in (normalized) MAD units from the median.
#define DEFAULT_AUTOSTRETCH_SCLIP  -2.80
// Target mean background in the [0,1] range.
#define DEFAULT_AUTOSTRETCH_TBGND   0.25
#define DEFAULT_AUTOSTRETCH_CLINK   true

#define Dispose "dispose"
#define tab '\t'
#define test 0
//
// csNRGB manages the channels, mask and mainview modification
//
function csNRGB(mainView, dialog)
{
	Console.writeln("mainView "+mainView.id);
	this.applied 	= false;
	this.copyName	= getNewName(mainView.id, "_copy");
	this.origView	= copyView( mainView, this.copyName);
	this.origView.window.hide();
	setDisposal(this.origView, true);
	this.mainView	= mainView;
	this.modifView	= null;
	this.dialog		= dialog;
	this.maskView	= null;
	this.useMask	= false;
	this.keepMask	= false;
	this.pixCount	= mainView.image.width * mainView.image.height;
	this.doRefresh	= false;
	this.channels	= [null, null, null];
	this.bitmap 	= linearViewToBMP(this.mainView);

	this.rgbPixels	= [];
	for (var i = 0; i < 3; i++)
	{
		var a       = new Array(this.pixCount);
		this.mainView.image.selectedChannel = i;
		this.mainView.image.getPixels(a);
		this.rgbPixels.push( new Float32Array(a) );
	}
	var  b  			= new ArrayBuffer(this.pixCount);
	this.maskFlags	= new Int8Array(b);		// mask map with 1
	for (var j = 0; j < this.pixCount; j++) this.maskFlags[j] = 1;

	this.channels	= extractChannels(mainView);
	for (var i in this.channels)
	{
		this.channels[i].window.hide();
		setDisposal(this.channels[i], true);
	}

	this.apply = function()
	{
		var description = "Emission Line Integration";
		description 	+= '\n';
		description 	+= "========================="
		description 	+= '\n';
		description 	+= this.mainView.id;
		description 	+= '\n';
		description 	+= this.channels[0].getHistory();
		description 	+= this.channels[1].getHistory();
		description 	+= this.channels[2].getHistory();
		if (this.useMask)
		{
			description += "Masked with " + this.maskView.id + '\n';
		}
		var P = new PixelMath;
		P.expression = this.modifView.id;
		P.expression1 = "";
		P.expression2 = "";
		P.expression3 = "";
		P.useSingleExpression = true;
		P.symbols = "";
		P.generateOutput = true;
		P.singleThreaded = false;
		P.use64BitWorkingImage = false;
		P.rescale = false;
		P.rescaleLower = 0;
		P.rescaleUpper = 1;
		P.truncate = true;
		P.truncateLower = 0;
		P.truncateUpper = 1;
		P.createNewImage = false;
		P.setDescription(description);
		if (P.executeOn(this.mainView))
		{
			this.applied = true;
			return true;
		}
		else
		{
			return false;
		}
	}

	this.refresh = function()
	{
		if (this.applied)
		{
			var P = new PixelMath;
			P.expression = this.origView.id;
			P.expression1 = "";
			P.expression2 = "";
			P.expression3 = "";
			P.useSingleExpression = true;
			P.symbols = "";
			P.generateOutput = true;
			P.singleThreaded = false;
			P.use64BitWorkingImage = false;
			P.rescale = false;
			P.truncate = false;
			P.createNewImage = false;
			P.executeOn(this.mainView)
		}
		this.applied = false;

		if (this.refreshed && this.doRefresh)
		{
			this.refreshed.call(this, this.dialog);
			this.doRefresh = false;
		}
	}

	this.setMask = function(maskView)
	{
		if (maskView.id != "")
		{
			if (!maskView.image.isGrayscale)
			{
				message(maskView.id + " is not a gray scale image");
				return;
			}
			else
			{
				var c = maskView.image.width * maskView.image.height;
				if ((maskView.image.width != this.mainView.image.width) ||
					 (maskView.image.height != this.mainView.image.height))
				{
					message(maskView.id + " has a different image geometry");
					return;
				}
			}


			var maskPix = new Array(this.pixCount);
			maskView.image.getPixels (maskPix);
			//
			// calculate dark areas, skip bright stars
			//
			var cnt = 0;
			var threshold = maskView.image.mean() + 1 * maskView.image.stdDev();
			for (var j = 0; j < this.pixCount; j++)
			{
				if (maskPix[j] > threshold)
				{
					this.maskFlags[j] = 0;		// skip star
					cnt += 1;
				}
				else
				{
					this.maskFlags[j] = 1;		// evaluate object & background
				}
			}
			// evaluate each pixel in this mask
			// set false in each location of MM for pixels
			// falling below a threshold
			//
			Console.writeln("# of pixel rejected " + cnt + ", mask threshold " +
								 threshold.toFixed(4));
			this.maskView   = maskView;
			this.useMask	= true;
		}
		else
		{
			this.useMask	= false;
			for (var j = 0; j < this.pixCount; j++) this.maskFlags[j] = 1;
		}
		//
		// setting/removing mask forces re-evaluation of mue
		//
		for (var i in this.channels)
		{
			this.channels[i].mue = 0.0;
		}
	}
	//
	//
	//
	this.addEmission = function()
	{
		var id = getNewName(this.mainView.id, "_Test");
		this.modifView = copyView(this.origView, id);
		//
		// mask modifView before PixelMath
		//
		if (this.useMask)
		{
			this.modifView.window.mask = this.maskView.window;
			this.modifView.window.maskEnabled = true;
			this.modifView.window.maskInverted = true;
		}

		var P = new PixelMath;
		P.expression 	= this.channels[0].expression();
		P.expression1	= this.channels[1].expression();
		P.expression2 	= this.channels[2].expression();
		P.useSingleExpression = false;
		P.symbols = "";
		P.generateOutput = true;
		P.singleThreaded = false;
		P.use64BitWorkingImage = false;
		P.rescale = false;
		P.rescaleLower = 0;
		P.rescaleUpper = 1;
		P.truncate = true;
		P.truncateLower = 0;
		P.truncateUpper = 1;
		P.createNewImage = false;
		var rc = false;
		try
		{
			rc =  P.executeOn( this.modifView );
		}
		catch (e)
		{
			rc = false
		}

		if (this.usemask)
		{
			this.modifView.window.removemask();
		}

		if (rc)
		{
			this.bitmap = linearViewToBMP(this.modifView);
			setDisposal(this.modifView, true);
			return this.modifView;
		}
		else
			return null;
	}

	this.setMaskDisposal = function(keep)
	{
		this.keepMask = keep;
		if (this.maskView != null)
		{
			setDisposal(this.maskView, !keep);
		}
	}


	this.maskShow = function(keep)
	{
		if (this.maskView != null)
		{
			var dispose = this.maskView.propertyValue(Dispose);
			if (!dispose)
			{
				this.maskView.window.show();
			}
		}
	}


	this.graphicalAnalysis = function()
	{
		// combine all graphs to one
		var images = [];

		for (var i = 0; i < 3; i++)
		{
			var channel = this.channels[i];
			if (channel.mue == 0) continue;
			var img1 = plotImage(channel.xy, channel.dxy, channel.mue,
							"Graphical analysis - black counts(mue)", i, this.pixCount);
			images.push(img1);
 		}
		//
		// assemble all to one view
		//
		var x = 0;
		var y = 0;
		var rows = images.length;
		if (rows > 0)
		{
			var sumImage = new Image(images[0].width * rows,
											 images[0].height,
											 3,
											 ColorSpace_RGB);
			for (var i = 0; i < images.length; i++)
			{
				var img = images[i];
				sumImage.apply( img, ImageOp_Add, new Point(x, y));
				x += img.width;
			}
			var window = new ImageWindow(sumImage.width, sumImage.height, 3);
			window.mainView.beginProcess();
			window.mainView.image.assign(sumImage);
			window.mainView.endProcess();
			window.show();
		}
	}

	function plotImage(xy, dxy, mue, title, channel, pixCount)
	{
		//
		// xy is black pixel count over mue
		// dxy is the 1st derivate
		//
		// create an image
		var w = 600;
		var h = 500;
		var cw = new Brush( 0xffFFFFFF);
		var cy = new Brush( 0xffCFCF00);
		var tickLen = 10;

		var colorXY  = 0xFF00FF00;
		var colorDXY = 0xFF0000FF;


		var y = 0;
		var x = 0;

		var maxX = 0;
		var maxY = 0;
		var minX = 1e+30;
		var minY = minX;

		var minDY = 1e+30;
		var maxDY = 0;
		//
		// transform y to percentage values
		//
		for (var i = 0; i < xy.length; i++)
		{
			xy[i][1] = xy[i][1] / pixCount * 100 ;
		}
		//
		// yOfMue, percentage of black at x = mue
		//
		var yOfMue = 0;
		for (var i = 1; i < xy.length; i++)
		{
			if (mue >= xy[i - 1][0] && mue <= xy[i][0])
			{
				// interpolate
				var dx = xy[i][0] - xy[i - 1][0];
				var dy = xy[i][1] - xy[i - 1][1];
				var s  = dy / dx;
				yOfMue = xy[i - 1][1] + (mue - xy[i - 1][0]) * s;
				break;
			}
		}
		//
		// minX -> maxX spans mue values of both curves
		// minDY - > maxDY is the range of y in dxy
		//
		for (var i = 0; i < xy.length; i++)
		{
			if (xy[i][0] > maxX) maxX = xy[i][0];
			if (xy[i][1] > maxY) maxY = xy[i][1];
			if (xy[i][0] < minX) minX = xy[i][0];
			if (xy[i][1] < minY) minY = xy[i][1];
		}

		for (var i = 0; i < dxy.length; i++)
		{
			if (dxy[i][0] > maxX) maxX = dxy[i][0];
			if (dxy[i][0] < minX) minX = dxy[i][0];

			if (dxy[i][1] > maxDY) maxDY = dxy[i][1];
			if (dxy[i][1] < minDY) minDY = dxy[i][1];
		}

		var bm  = new Bitmap(w, h);

		var g   = new Graphics(bm);

		g.fillRect(0, 0, bm.width, bm.height, cw);
		g.drawRect(0, 0, bm.width, bm.height);

		//g.antialiasing = true;

		g.font = new Font( FontFamily_SansSerif, 8);
		// -------------------------------------------------------------------------
		//
		// draw scales and text
		//
		// -------------------------------------------------------------------------
		var yLabelRect = new Rect(0, 0, w, h);		// space for 4 digits
		yLabelRect = g.textRect( yLabelRect, "1234", TextAlign_Left);
		// -------------------------------------------------------------------------
		// the curve drawing frame
		// -------------------------------------------------------------------------
		var rect = new Rect(yLabelRect.x1 + tickLen + 1, 40, w - 30, h - 40);
		// -------------------------------------------------------------------------
		// find optimal scaling in x and y-axis
		// -------------------------------------------------------------------------
		var ticks = 10;
		var soX = axisScaling(minX, maxX, ticks, rect.width);
		var soY = axisScaling(minY, maxY, ticks, rect.height);

		var xScale = soX.axisScale;
		var yScale = soY.axisScale;
		//
		// scale y (dxy) from 0 -> plot rectangle height * 0.99
		//
		var dScale = rect.height * 0.99 / (maxDY - minDY);

		var ticksX = soX.roundedTickRange;
		var ticksY = soY.roundedTickRange;

		var x0 = Math.floor(minX / ticksX) * ticksX;
		var y0 = Math.floor(minY / ticksY) * ticksY;
		// -------------------------------------------------------------------------
		// transform data points xy
		// -------------------------------------------------------------------------
		var pointsXY = [];
		var pointsDXY = [];
		for (var i = 0; i < xy.length; i++)
		{
			pointsXY.push(new Point((xy[i][0] - x0) * xScale,
											(xy[i][1] - y0) * yScale));
		}
		yOfMue -= y0;
				Console.writeln("yOfMue "+yOfMue);
		yOfMue *= yScale;
				Console.writeln("yOfMue "+yOfMue);

		for (var i = 0; i < dxy.length; i++)
		{
			pointsDXY.push(new Point((dxy[i][0] - x0) * xScale,
											 (dxy[i][1] - minDY) * dScale));
 		}
		// -------------------------------------------------------------------------
		// X-Scale in bottom rect
		// -------------------------------------------------------------------------
		g.clipRect = new Rect(0, rect.y1, w, h);

		var xLabelRect = g.textRect( g.clipRect, ticksX.toString(), TextAlign_Left);

		x = rect.x0;
		v = x0;
		for (;;)
		{
			g.antialiasing = true;
			g.drawLine(g.clipRect.x0 + x,
						  g.clipRect.y0,
						  g.clipRect.x0 + x,
						  g.clipRect.y0 + tickLen);
			xLabelRect.center = new Point(x, g.clipRect.y0 + tickLen +
															xLabelRect.height / 2);
			g.antialiasing = false;
			g.drawTextRect(xLabelRect, Math.roundTo(v, 6).toString(),
															TextAlign_Center);
			v += ticksX;
			x += ticksX * xScale;
			if (x > rect.x1) break;
		}
		// -------------------------------------------------------------------------
		// Y-Scale in left rect
		// -------------------------------------------------------------------------
		g.clipRect = new Rect(0, 0, rect.x0, h);
		var labelHH = yLabelRect.height / 2;
		y = rect.y1;
		var L = Math.floor(Math.log10(ticksY));	// int log of unit
		L     = Math.floor(L / 3) * 3;
		var P = Math.pow(10, L);

		var v = y0;

		for (;;)
		{
			g.antialiasing = true;
			g.drawLine(g.clipRect.x1 - tickLen, y, g.clipRect.x1, y);
			yLabelRect = new Rect(0, y - labelHH, yLabelRect.x1, y + labelHH);
			g.antialiasing = false;
			g.drawTextRect(yLabelRect, (v/P).toString(), TextAlign_Right |
																		TextAlign_VertCenter);
			v += ticksY;
			y -= ticksY * yScale;
			if (y < rect.y0) break;
		}

		var lStr = L.toString();	// log, format [10p]

		yLabelRect = g.textRect( new Rect(tickLen, tickLen, w, rect.y0), "[%]",
														TextAlign_Left );
		g.drawTextRect(yLabelRect, "[%]");

		g.font = new Font( FontFamily_SansSerif, 8);

		var numFont = g.font;	// save Font
		// -------------------------------------------------------------------------
		// Title
		// -------------------------------------------------------------------------
		g.font = new Font( FontFamily_Times, 14);
		g.font.bold = true;

		g.clipRect = new Rect(0, 0, w, h);
		var txtR = new Rect(rect.x0, 0, rect.x1, rect.y0);

		txtR = g.textRect( txtR, title, TextAlign_Center);
		g.drawTextRect( txtR, title);

		var channelStr = "Channel " + "RGB"[channel];

		g.font = new Font( FontFamily_Times, 10);
		g.font.bold = false;
		txtR = g.textRect(new Rect(0, 0, rect.x1, rect.y0),
											channelStr,
											TextAlign_Right | TextAlign_VertCenter);
		g.drawTextRect( txtR, channelStr);

		// draw legend (the height of current font is measured with txtR)

		g.font = new Font( FontFamily_SansSerif, 8);

		var charHeight = g.textRect(rect, "X").height ;

		y = rect.y0 + 15;

		g.pen = new Pen(colorXY);
		g.pen.width = 5;
		g.drawLine(rect.x0 + 4, y, rect.x0 + 32, y);

		g.pen = new Pen(0xFF000000);
		g.drawText(rect.x0 + 34, y + charHeight / 3,
						"percentage of black pixel");

		y += charHeight;

		g.pen = new Pen(colorDXY);
		g.pen.width = 5;
		g.drawLine(rect.x0 + 4, y, rect.x0 + 32, y);

		g.pen = new Pen(0xFF000000);
		g.drawText(rect.x0 + 34, y + charHeight / 3, "1st derivative");

		// draw curve in transformed box

		g.clipRect = rect;

		g.scaleTransformation( 1, -1 );
		g.translateTransformation( rect.x0, -rect.y1 );
		g.antialiasing = true;

		g.pen = new Pen(colorXY);
		g.pen.width = 3;

		g.drawPolyline( pointsXY );

		g.pen = new Pen(colorDXY);
		g.pen.width = 3;

		g.drawPolyline( pointsDXY );

		g.antialiasing = false;
		g.pen = new Pen(0xFF8F0000);
		g.drawLine(0, yOfMue, rect.width, yOfMue);

		g.resetTransformation();

		g.pen = new Pen(0xFF000000);
		g.pen.width = 1;

		g.clipRect = new Rect(0, 0, w, h);

		g.antialiasing = true;
		g.drawRect( rect.x0, rect.y0, rect.x1 , rect.y1 );
		// -------------------------------------------------------------------------
		// draw text: m = mue with symbol for m
		// -------------------------------------------------------------------------
		var mueP0 = new Point(rect.x0 + (mue -x0) * xScale, rect.y0);
		var mueP1 = new Point(rect.x0 + (mue -x0) * xScale, rect.y1);

		var mueStr = "m = " + mue.toString();

		g.font = new Font( FontFamily_Symbol, 14);
		g.font.bold = true;
		var mueFont = g.font;
		var mRectM = g.textRect(rect, "m",
										TextAlign_Right | TextAlign_Bottom);

		g.font = numFont;
		var mRectN = g.textRect(rect, "= " + mue.toFixed(6),
										TextAlign_Right | TextAlign_Bottom);

		mRectM = new Rect(mRectN.left, mRectM.top,
								mRectN.left - (mRectM.x1 - mRectM.x0),
								mRectM.bottom);

		mRectM.add(-6, -6);
		mRectN.add(-6, -6);
		// -------------------------------------------------------------------------
		//
		// -------------------------------------------------------------------------
		g.font = mueFont;

		g.drawTextRect(mRectM, "m");

		g.font = numFont;

		g.drawTextRect(mRectN, "= " + mue.toFixed(6));
		// -------------------------------------------------------------------------
		// draw red line Mue
		// -------------------------------------------------------------------------
		g.antialiasing = true;
		g.pen = new Pen(0xFFFF0000);
		g.drawLine(mueP0, mueP1);
		// -------------------------------------------------------------------------
		// close graphics
		// -------------------------------------------------------------------------
		g.end();
		// -------------------------------------------------------------------------
		// bitmap to image
		// -------------------------------------------------------------------------
		var img = new Image(bm.width, bm.height, 3, ColorSpace_RGB);
		img.blend(bm);

		return img;
	}


	function axisScaling(minX, maxX, ticks, axisLen)
	{
		var range = maxX - minX;
		var unroundedTickSize = range / (ticks - 1);
		var x = Math.ceil(Math.log10(unroundedTickSize) - 1);
		var pow10x = Math.pow(10, x);
		var roundedTickRange = Math.ceil(unroundedTickSize / pow10x) * pow10x;
		var n = Math.floor(range / roundedTickRange) + 1;
		var axisScale = axisLen / (n * roundedTickRange);
		return {"roundedTickRange":roundedTickRange, "axisScale":axisScale};
	}

	function copyView( view, newName)
	{
		var P = new PixelMath;
		P.expression = "$T";
		P.expression1 = "";
		P.expression2 = "";
		P.expression3 = "";
		P.useSingleExpression = true;
		P.symbols = "";
		P.generateOutput = true;
		P.singleThreaded = false;
		P.use64BitWorkingImage = false;
		P.rescale = false;
		P.rescaleLower = 0;
		P.rescaleUpper = 1;
		P.truncate = true;
		P.truncateLower = 0;
		P.truncateUpper = 1;
		P.createNewImage = true;
		P.showNewImage = false;// true;
		P.newImageId = newName;
		P.newImageWidth = 0;
		P.newImageHeight = 0;
		P.newImageAlpha = false;
		P.newImageColorSpace = PixelMath.prototype.SameAsTarget;
		P.newImageSampleFormat = PixelMath.prototype.SameAsTarget;
		if (P.executeOn(view))
			return View.viewById(newName);
		{
			message("Mask creation failed in function copyView", "Error");
			return null;
		}
	}

}
//
// csChannel, class with methods and properties for one
// 	braodband/narrowband combination
//
function csChannel(NRGB, channelIndex)
{
	this.NRGB				= NRGB;
	this.channelIndex		= channelIndex;
	this.broadBand 		= this.NRGB.channels[channelIndex];
	this.narrowBand 		= null;
	this.emission 			= null;
	this.mue 				= 0.0;
	this.dialog				= this.NRGB.dialog;
	this.mueCorrection	= 1.0;
	this.contingent		= 100.0;
	this.medBroadband		= this.broadBand.image.median();
	this.medNarrowband	= 0.0;
	this.hasNarrowBand	= false
	this.nbPixelArray		= null;
	this.bitmap				= null;
	this.maskApplied		= false;
	this.xy     			= null;						// mue, count pairs
	this.dxy    			= null;						// dx/dy (this.xy)

	//
	// change views to csChannel class
	//
	NRGB.channels[channelIndex] = this;
  	//
	//
	//
	this.setNarrowBand = function(narrowBand)
	{
		this.maskApplied = false;
		this.narrowBand  = narrowBand;
		this.mue = 0.0;
		if (this.narrowBand.id != "")
		{
			var a       		 = new Array(this.NRGB.pixCount);
			this.narrowBand.image.getPixels(a);
			this.nbPixelArray	 = new Float32Array(a);
			this.medNarrowband = narrowBand.image.median();
			this.hasNarrowBand = true;
		}
		else
			this.hasNarrowBand = false;
		this.NRGB.refresh();
		if (this.emission != null)
		{
			this.emission.window.forceClose();
		}
	}
 	//
	//
	//
	this.setMueCorrection = function(mueCorrection)
	{
		this.mueCorrection = mueCorrection;
	}
 	//
	//
	//
	this.setContingent = function(contingent)
	{
		this.contingent = contingent;
	}
 	//
	//
	//
	this.getEmission = function()
	{
		if (this.narrowBand != null)
		{
			if (this.mue == 0.0)
			{
				//
				//
				//
				if (this.emission != null)
				{
					this.emission.window.forceClose();
				}
				//
				var initialMue = this.medNarrowband / this.medBroadband * 0.5;
				var curves = new mueCurves(this.nbPixelArray,
											  		this.NRGB.rgbPixels[this.channelIndex],
										  	  		this.NRGB.maskFlags,
										  	  		initialMue,
										  	  		this.dialog);
				this.mue = curves.mue;
				this.xy  = curves.xy;
				this.dxy = curves.dxy;
			}
			if (this.mue > 0.0)
			{
				this.emission = continuumSubtract(this.narrowBand,
															 this.broadBand,
 															 this.mue * this.mueCorrection,
 															 this.channelIndex);
 				this.emission.window.hide();
		 		this.bitmap = linearViewToBMP(this.emission);
				setDisposal(this.emission, true);
				this.NRGB.doRefresh = true;
				//
				// call back, when emission calculation is done
				//
				this.csSubtracted.call(this, this);
				return true;
			}
				return false;
		}
		else
		{
			return true;
		}
	}
	//
	//
	//
	this.expression = function()
	{
 		if (this.mue == 0.0)
		{
			return "$T";
		}
		else
		{
			return "$T + " + this.emission.id + " * " + this.contingent * 0.01;
		}
	}
	//
	// create a history string
	//
	this.getHistory = function()
	{
		var s = "Channel " + this.channelIndex;
		if (this.mue == 0.0)
		{
			s += " no modification" + '\n';
		}
		else
		{
			s += '\n';
			s += "Narrowband " + this.narrowBand.id + '\n';
			s += "Mue " + this.mue;
			if (this.mueCorrection != 1.0)
				s += ", correction " + this.mueCorrection;
			s += '\n';
			s += "Contingent " + this.contingent + '\n';
		}
		return s;
	}

	/* ***************************************************************************
	 *
	 * Evaluate the optimal quantity to be subtracted from the narrow band channel
	 *
	 * This quantity is the broad band image x mue
	 *
	 * ***************************************************************************/
	function mueCurves(NB, BB, MM, initialMue, dlg)
	{
		//
		// evaluate first derivative of counts over mue
		//
		const minStep = 0.00025;
		var stp     = minStep;
		var n       = NB.length;
		var fx		= [];
		var bp 		= new Int8Array(MM);	// blackpoints
		var ms 		= 0;						// max. slope
		var im   	= 0;						// index of max
		var mueM 	= initialMue * 5;
		var i 		= -1;						// loop index
		var stps 	= stp * 12.0;			// slope base
		var lim		= 10 / (minStep * 4);

		this.mue 	= initialMue - stp;
		this.xy     = [];						// mue, count pairs
		this.dxy    = [];						// dx/dy (this.xy)

		for (;;)
		{
			i += 1;							// 0..

			this.mue += stp;

			advanceProgress(dlg);

			processEvents();

			if (dlg.cancelPending) return -1;

			if(this.mue > mueM)
			{
				// no solution
				this.mue = -this.mue;
				break;
			}
			// -----------------------------------------------------------------------
			// subtract mue-fraction of BB
			// -----------------------------------------------------------------------
			var cnt    = 0;
			for (var j = 0; j < n; j++)
			{
				if (bp[j])					// if 1, check this position
				{
					//
					// count non-masked pixels
					//
					var v = NB[j] - BB[j] * this.mue;
					if (v > 0) continue;
					cnt += 1;
					bp[j] = 0;				// skip this pixel in next run
				}
				else
					cnt += 1;
			}

			fx.push(cnt);
			this.xy.push([this.mue, cnt]);
			// -----------------------------------------------------------------------
			// -----------------------------------------------------------------------
			// -----------------------------------------------------------------------
			if (i < 5) continue;

			var slope = (fx[i - 4] - 8 * fx[i - 3] + 8 * fx[i - 1] - fx[i]) / stps;

			if (Math.abs(slope < lim)) slope = 0;

			this.dxy.push([this.mue, slope]);

			if (slope > 0)
			{
				if (slope > ms)
				{
					ms  = slope;
					im  = this.dxy.length - 1;		// keep index of maximum
				}
				else if (im > 0 && ((this.dxy.length - im) > 50))
				{
					//
					// curve falling
					//
					var i0  = Math.max(0, im - 5);
					var i1  = Math.min(this.dxy.length - 1, im + 5);
					var par = new lstqr();
					for (var j = i0; j < i1; j++)
					{
						par.addPoint(this.dxy[j][0], this.dxy[j][1]);
					}
					par.solve();
					this.mue = par.XS();
					break;
				}
			}
			else
			{
				ms = 0.0;
			}
		}
	}

	function continuumSubtract(nb, bb, mue, channelIndex)
	{
		var name = getNewName("emission", "_" + "RGB".charAt(channelIndex));
		var P = new PixelMath;
		P.expression = nb.id + " - " + bb.id + " * " + mue;
		P.expression1 = "";
		P.expression2 = "";
		P.expression3 = "";
		P.useSingleExpression = true;
		P.symbols = "";
		P.generateOutput = true;
		P.singleThreaded = false;
		P.use64BitWorkingImage = false;
		P.rescale = false;
		P.rescaleLower = 0;
		P.rescaleUpper = 1;
		P.truncate = true;
		P.truncateLower = 0;
		P.truncateUpper = 1;
		P.createNewImage = true;
		P.showNewImage = false;
		P.newImageId = name;
		P.newImageWidth = nb.image.width;
		P.newImageHeight = nb.image.height;
		P.newImageAlpha = false;
		P.newImageColorSpace = PixelMath.prototype.Gray;
		P.newImageSampleFormat = PixelMath.prototype.f32;
		if (P.executeGlobal())
		{
			return View.viewById(name);
		}
		else
		{
			return null;
		}
	}
}

/* *****************************************************************************
 * *****************************************************************************
 *		DIALOG
 * *****************************************************************************
 * *****************************************************************************
 * */

function showDialog(viewRGB)
{
   // Add all properties and methods of the core Dialog object to this object.
   this.__base__ = Dialog;
   this.__base__();
   this.userResizable = true;

   var RGB = "RGBM"
   var picturebox;

   this.cancelPending = false;

	var nbTested = false;

   // control arrays

   var ctls    				= [null, null, null];
   var c_views 				= [null, null, null];
	this.contingentControls = [null, null, null];
	this.mueCorrectControls = [null, null, null];
	this.greenLabelControls = [null, null, null];
	this.radioButtons 		= [null, null, null, null];

	//

	this.NRGB = new csNRGB(viewRGB, this);
	this.NRGB.refreshed = function (dialog)
	{
		if (dialog.radioButtons[0].checked)
		{
			dialog.previewControl.SetImage( dialog.NRGB.bitmap, dialog.data);
		}
		dialog.applyButton.enabled = false;
	}

	this.csChannels = [null, null, null];
	for (var i = 0; i < 3; i++)
	{
		this.csChannels[i] = new csChannel(this.NRGB, i);
		with (this.csChannels[i])
		{
			this.csChannels[i].csSubtracted = function(channel)
			{
				with (channel)
				{
					//
					// is this channel selected by radioButton?
					//
					dialog.radioButtons[channelIndex + 1].enabled = true;
				 	for (var i = 0; i < 3; i++)
					{
						if (dialog.radioButtons[i + 1].checked)
						{
							dialog.previewControl.SetImage(bitmap, dialog.data);
						}
					}
				}
			}
		}
	}

   this.progress = 0;

	var tempName = getNewName(viewRGB.id, "_temp");
	//
	// linear view
	//
	var view = copyView(viewRGB, tempName);
	view.setPropertyValue( Dispose, true )
	view.window.hide();

	// ------------------------------------------------------------------------
   // Preview
   // ------------------------------------------------------------------------

	this.data = new imageMetadata(view);
	this.previewControl = new PreviewControl(this);
	this.previewControl.stf_Button.visible = false;
	this.onCustomSTF = function()
	{
		//
	}

   //

   console.writeln("Narrowband application on RGB images");
   console.writeln("====================================");
   console.writeln("Processing " + viewRGB.id);

	// ------------------------------------------------------------------------
   // operating controls
   // ------------------------------------------------------------------------

   this.lblHeadLine = new Label(this)
   with (this.lblHeadLine)
   {
      useRichText = true;
      text ="<b>RGB: " + view.id +"</b>";
   }

   // my ©


   this.lblCopyright = new Label(this)
   this.lblCopyright.text = "© 2017, Hartmut V. Bornemann";

	for (var i = 0; i < 4; i++)
	{
		var radioButton = new RadioButton(this);
		with (radioButton)
		{
			if (i == 0)
			{
				checked = true;
				text = "RGB";
				toolTip = "View the current RGB image";
			}
			else
			{
				text = "E" + i;
				enabled = false;
				toolTip = "View the emission line image";
			}
			onCheck = function( checked )
			{
				if (parent == null) return;
				for (var j = 0; j < parent.radioButtons.length; j++)
				{
					if (parent.radioButtons[j] == this)
					{
						if (j > 0)
						{
							//
							// view RGB
							//
							parent.previewControl.SetImage(
								parent.csChannels[j - 1].bitmap,	parent.data);
						}
						else
						{
							//
							// view narrowband
							//
							parent.previewControl.SetImage(parent.NRGB.bitmap,
								parent.data);
						}
					}
				}
			}
		}
		this.radioButtons[i] = radioButton;
	}

	this.testButton =  new PushButton( this );
	with (this.testButton)
	{
		enabled = false;
      defaultButton = true;
		text = "Test";
		toolTip = "<p>Evaluate subtraction quantity for each channel" +
					 " and display the result in the preview</p>";
		onClick = function(checked) {
			enabled = false;
			parent.okButton.enabled = false;
			parent.applyButton.enabled = false;
			parent.progress = 0;

			this.parent.NRGB.refresh();

			for (var i in this.parent.csChannels)
			{
				parent.greenLabelControls[i].foregroundColor = 0xFF00FF00;

				if (this.parent.csChannels[i].getEmission())
				{
					with (this.parent.csChannels[i])
					{
						if (mue > 0.0)
						{
							Console.writeln(emission.id + " = " + narrowBand.id +
							" - " + mue.toFixed(5) + " * " + broadBand.id);
						}
					}
					parent.greenLabelControls[i].foregroundColor = 0x40808080;
					continue;
				}
				Console.writeln("get emission function failed");
				return;
			}

			this.parent.NRGB.addEmission();
			//view = this.parent.NRGB.addEmission();
			if (dialog.radioButtons[0].checked)
			{
				this.parent.previewControl.SetImage(this.parent.NRGB.bitmap,
																this.parent.data);
			}

			enabled = true;
			nbTested = true;
			parent.okButton.enabled = true;
			parent.applyButton.enabled = true;
			parent.progress = 0;
			advanceProgress(parent);
		}
	}

	this.applyButton = new PushButton( this );
	with (this.applyButton)
	{
      defaultButton = false;
		text = "Apply";
		toolTip = "<p>Apply the reduced NB images to the MainView</p>";
		enabled = false;
		onClick = function(checked) {
			if (this.parent.NRGB.apply())
			{
				stfMainView(parent);
			}
			else
				Console.writeln("Update failed");
			enabled = false;
			nbTested = false;
		}
	}


   this.okButton = new PushButton( this );
   with (this.okButton)
   {
      defaultButton = false;
      icon = this.scaledResource( ":/icons/ok.png" );
      text = "OK";
		toolTip = "<p>Exit this dialog</p>";

      onClick = function(checked)
      {
			if (nbTested)
			{
				var query = new MessageBox("The image was modified'\r\n'"
							+ "'Yes' copies to view '\r\n'"
							+ "'Cancel' continues '\r\n'"
							+ "'No' quits without updating",
							"Save view",
											StdIcon_Question,
											StdButton_Yes,
											StdButton_No,
											StdButton_Cancel);
				var answ = query.execute();
				if (answ == StdButton_Yes)
				{
					if (parent.NRGB.apply())
					{
						stfMainView(parent);
					}
					else
					{
						Console.writeln("Update failed");
					}
				}
				else if (answ == StdButton_Cancel)
				{
					return; // continue script
				}
				else
				{
					Console.writeln("Image not updated");
				}
			}
         //
			// show or remove views
			//
			if (dialog.retainNB.checked)
			{
 		 		keepEmissionLineViews(parent);
				stfEmissionViews(parent);
				//
				// add the graphical analysis
				//
				parent.NRGB.graphicalAnalysis();
			}
			//
			//
			//
			parent.NRGB.maskShow();
			//
			// current parameters to Settings
			//
			for (var i = 0; i < 3; i++)
			{
				Settings.write("ELIMueCorrection_" + i, DataType_Double,
									parent.csChannels[i].mueCorrection );
				Settings.write("ELIContingent_" + i, DataType_Double,
									parent.csChannels[i].contingent );
			}
			//
			// write files in test-mode
			//
			if (test)
			{
				for (var i = 0; i < 3; i++)
				{
					if (this.parent.csChannels[i].xy != null)
					{
						var filename = File.systemTempDirectory +
											"\\XYEmission" + "RGB"[i] + ".csv";
						var f = new File(filename,FileMode_Create);
						f.outTextLn( "mue" +tab + "count")
						for (var j = 0; j < this.parent.csChannels[i].xy.length; j++)
						{
							f.outTextLn( this.parent.csChannels[i].xy[j][0] + tab +
											 this.parent.csChannels[i].xy[j][1])
						}
						f.close();
					}
					if (this.parent.csChannels[i].xy != null)
					{
						var filename = File.systemTempDirectory +
											"\\DXYEmission" + "RGB"[i] + ".csv";
						var f = new File(filename,FileMode_Create);
						f.outTextLn( "mue" +tab + "count")
						for (var j = 0; j < this.parent.csChannels[i].dxy.length; j++)
						{
							f.outTextLn( this.parent.csChannels[i].dxy[j][0] + tab +
											 this.parent.csChannels[i].dxy[j][1])
						}
						f.close();
					}
				}
			}
			//
			// end dialog
			//
			parent.done(0);
      };
   }
   //
   // create a controlbox for each channel and one optional mask
   //
   for (var i = 0; i < 4; i++)
   {
      var gbx = new GroupBox( this );
      with (gbx)
      {
         sizer = new HorizontalSizer;
         sizer.margin = 6;
         sizer.spacing = 4;
         if (i < 3)
			{
            title = "Channel " + RGB[i];
			}
         else
			{
            title = "Star Mask [opt.]";
			}
      }
      ctls[i] = gbx;

      // label: Narrowband image

      var lblImage = new Label( this );
      with (lblImage)
      {
         if (i < 3)
         {
            text = "Narrowband image:";
            toolTip = "<p>Narrowband image to be added to channel " +
                     RGB[i] + ".</p>";
				setFixedWidth(width);
         }
         else
         {
            text = "Star mask:";
            toolTip = "<p>Exclude stars with this optional binary star mask.</p>";
        	}
         textAlignment = TextAlign_Left|TextAlign_VertCenter;
      }

      // image select

		var vList = new ViewList( this );
      with (vList)
      {
	   	objectId = "view+" + i;
			setVariableSize();
         getAll();
         onViewSelected = function( view )
			{
         	this.cursor = new Cursor(StdCursor_Wait);
				var i = ctlIndex(this);
				if (i < 3)
				{
					parent.parent.csChannels[i].setNarrowBand(view);
					//
					// set controls with mue, mueCorrection, contingent from view
					//
					with (parent.parent.csChannels[i])
					{
						parent.parent.radioButtons[i + 1].enabled = hasNarrowBand &&
						emission != null;;
					}
				}
				else
				{
					//
					// set mask
					//
					parent.parent.NRGB.setMask(view);
				}
         	this.cursor = new Cursor(StdCursor_Arrow);
				//
				// at least one channel selected?
				//
				parent.parent.testButton.enabled = false;
				for (var j in parent.parent.csChannels)
				{
					if (parent.parent.csChannels[j].hasNarrowBand)
					{
						parent.parent.testButton.enabled = true;
						break;
					}
				}
			}
		}
		//
		// label greenLight
		//
		if (i < 3)
		{
			var lblGreenLight  = new Label(this);
			lblGreenLight.font = new Font("WebDings", 11)
			lblGreenLight.text = "n";
			lblGreenLight.foregroundColor = 0x40808080;
			lblGreenLight.setFixedWidth( lblGreenLight.height );
			this.greenLabelControls[i] = lblGreenLight;
		}
      // label: Contingent
		//
      var numContingent;
      var lblMueCorrection;
      var trkMueCorrection;
      if (i < 3)
      {
         numContingent = new NumericEdit(this);
         with (numContingent)
         {
				autoEditWidth = true;
            objectId = "contingent+" + i;
            label.text = "Contingent:";
            setPrecision( 0 );
            setRange( 0, 999 )
            setValue( 100 )
            toolTip = "<p>Choose a percentage contingent, if this narrow band " +
                      "channel is added also to another one.</p>";
            onValueUpdated = function( value )
            {
               var i = ctlIndex(this);
             	this.parent.parent.csChannels[i].setContingent(value);
            }
            onMouseDoubleClick = function( x, y, buttonState, modifiers )
            {
               setValue (100);
               var i = ctlIndex(this);
           		this.parent.parent.csChannels[i].setContingent(100);
            }
				onMouseWheel = function( x, y, delta, buttonState, modifiers )
				{
               var i = ctlIndex(this);
					var step = 10 * Math.sign(delta);
					var v = value - step;
					if (v > upperBound) v = upperBound;
					if (v < lowerBound) v = lowerBound;
					setValue(v);
             	this.parent.parent.csChannels[i].setContingent(v);
				}
         }
			this.contingentControls[i] = numContingent;
         //
         trkMueCorrection = new NumericControl (this);
         with ( trkMueCorrection )
         {
				autoEditWidth = true;
            objectId = "mueCorrection+" + i;
            label.text = "Mue correction:";
            slider.scaledMinWidth = 150;
            setReal( true );
            setPrecision (2);
            setRange( 0.50, 1.5 );
            slider.setRange( 500, 1500 );
            setValue (1);
            toolTip = "<p>Mue correction. Use this, if subtraction leads to" +
                      " under or over subtracted images.</p>";
            onValueUpdated = function( value )
            {
               var i = ctlIndex(this);
            	this.parent.parent.csChannels[i].setMueCorrection(value);
            }
            onMouseDoubleClick = function( x, y, buttonState, modifiers )
            {
               setValue (1);
               var i = ctlIndex(this);
         		this.parent.parent.csChannels[i].setMueCorrection(1);
            }
				onMouseWheel = function( x, y, delta, buttonState, modifiers )
				{
               var i = ctlIndex(this);
					var step = 0.1 * Math.sign(delta);
					var v = value - step;
					if (v > upperBound) v = upperBound;
					if (v < lowerBound) v = lowerBound;
					setValue(v);
             	this.parent.parent.csChannels[i].setMueCorrection(v);
				}
        	}
 			this.mueCorrectControls[i] = trkMueCorrection;
      }

      // fill the groupBox

		with (gbx.sizer)
      {
         spacing = 4;
         add( lblImage );
        	addSpacing( 8 );
         add( vList );
         if (i < 3)
         {
            addSpacing( 16 );
				add( lblGreenLight );
            addSpacing( 16 );
            add( numContingent );
            addSpacing( 16 );
            add( trkMueCorrection );
				lblGreenLight.setFixedWidth(lblGreenLight.width);
				numContingent.setFixedWidth(numContingent.width);
				trkMueCorrection.setFixedWidth(trkMueCorrection.width);
         }
			else
			{
				vList.setFixedWidth(trkMueCorrection.width);
			}
     	}

      // save controls

      c_views[i] = vList;
   }

   // CheckBoxes

   this.retainNB = new CheckBox(this);
   with (this.retainNB)
   {
      text = "Retain emission view(s)";
      toolTip = "<p>Keep emission line view for visual or scientific analysis.</p>";
   }

   // OK and Cancel button

   this.progressBar = new Frame(this);
   with (this.progressBar)
   {
      //visible = false;
      onPaint = function()
      {
         var g      = new Graphics();
         g.begin(picturebox);
         g.clipRect = new Rect(0, 0, width, height);
         if (parent.progress == 0)
         {
            var clear = new Brush( parent.backgroundColor );
            g.clipRect = new Rect(0, 0, width, height);
            g.fillRect( g.clipRect, clear );
         }
         else
         {
            var clear = new Brush( Color.WHITE );
            var r = new Rect(0, 0, parent.width - position.x * 2, height);
            g.clipRect = r;
            g.fillRect( g.clipRect, clear );
            var w = parent.progress * r.width * 0.01;
            r = new Rect(0, 0, w, height);
            g.fillRect(r, new Brush( Color.GREEN ));
         }
         g.end();
      }

      onResize = function(wNew, hNew, wOld, hOld)
      {
         resize(parent.width-16, 6);
         picturebox.repaint();
      }

      onShow = function()
      {
         resize(parent.width - 16, 6);
         picturebox.repaint();
      }
   }
   picturebox = this.progressBar;

	//
	// arrange controls in frames
	//

   this.pbarSz = new HorizontalSizer;
   with (this.pbarSz)
   {
      margin = 8;
      spacing = 8;
      add( this.progressBar );
   }

   this.buttons = new HorizontalSizer;
   with (this.buttons) {
      spacing = 8;
		add( this.testButton );
		addStretch();
		add( this.applyButton );
		addStretch();
      add( this.retainNB );
      add( this.okButton );
   }


	this.topFrameSizer =  new HorizontalSizer
   with ( this.topFrameSizer ) {
		this.previewControl.setMinSize(800, 500 );
		add( this.previewControl );
	}

	this.radioButtonSizer = new HorizontalSizer;
	with (this.radioButtonSizer)
	{
		add( this.lblCopyright );
		addStretch();
		for (var i = 0; i < this.radioButtons.length; i++)
		{
      	addSpacing( 16 );
			add(this.radioButtons[i]);
		}
	}

   this.bottomFrameSizer = new VerticalSizer;
   with ( this.bottomFrameSizer ) {
      margin = 6;
      spacing = 6;
      add( this.pbarSz );
      add( this.lblHeadLine );
		add(this.radioButtonSizer);
      addSpacing( 8 );
      for (var i = 0; i < ctls.length; i++)
      {
         add( ctls[i] );
      }
      add( this.buttons);
   }

   // all together
   this.sizer = new VerticalSizer;
   with ( this.sizer ) {
      margin = 8;
		add( this.topFrameSizer );
      spacing = 8;
      add( this.bottomFrameSizer );
   }

   // window
   this.windowTitle = TITLE + " " + VERSION;
   this.adjustToContents();

	processEvents();

	for (var i = 0; i < 3; i++)
	{
		var value = null;
		value = Settings.read("ELIMueCorrection_" + i, DataType_Double);
		if (value != null)
		{
			this.mueCorrectControls[i].setValue( value );
			this.csChannels[i].setMueCorrection( value );
		}
		value = Settings.read("ELIContingent_" + i, DataType_Double);
		if (value != null)
		{
			this.contingentControls[i].setValue( value );
			this.csChannels[i].setContingent( value );
		}
	}

	this.previewControl.SetImage(this.NRGB.bitmap, this.data);

   //

   function ctlIndex( ctl )
   {
      // get channel index of a control
      var items = ctl.objectId.split("+");
      if (items.length == 2) return items[1].toInt();
      return -1;
   }

	function getControl(objectId, parentControl)
	{
	}

	function protocol()
	{
		//
		// write protocol
		//
		if (this.NRGB.applied)
		{
			console.writeln(stringOf('=', 72));
			console.writeln("Add emission line(s) to broadband RGB");
			if (this.NRGB.useMask) console.writeln("Mask " + this.NRGB.maskView.id);
			console.writeln("");
			console.writeln("       | " + this.csChannels[0].expression);
			console.writeln(" RGB = | " + this.csChannels[1].expression);
			console.writeln("       | " + this.csChannels[2].expression);
			console.writeln("");
			Console.writeln(stringOf('=', 72));
		}
	}

	function stfMainView(dialog)
	{
		//
		// ApplyAutoSTF on image
		//
		ApplyAutoSTF( viewRGB,
					DEFAULT_AUTOSTRETCH_SCLIP,
					DEFAULT_AUTOSTRETCH_TBGND, true );
	}

	function stfEmissionViews(dialog)
	{
		//
		// ApplyAutoSTF emission line view(s)
		//
		for (var i = 0; i < 3; i++)
		{
			if (dialog.csChannels[i].emission != null)
			{
				if (dialog.retainNB.checked)
				{
					Console.writeln("Emission on " + RGB[i] + ", View " +
						dialog.csChannels[i].emission.id+ ", mue = " +
						dialog.csChannels[i].mue );
					//
					// retain emission line views?
					//
					ApplyAutoSTF(dialog.csChannels[i].emission,
							DEFAULT_AUTOSTRETCH_SCLIP,
							DEFAULT_AUTOSTRETCH_TBGND, true );
				}
			}
		}
	}

	function keepEmissionLineViews(dialog)
	{
		//
		// keep or retain emission line views depending on dislog's checkBox
		//
		for (var i = 0; i < 3; i++)
		{
			if (dialog.csChannels[i].emission != null)
			{
				//
				// keep away from disposal function
				//
				setDisposal(dialog.csChannels[i].emission, false);
				dialog.csChannels[i].emission.window.show();
			}
		}
	}
}


function advanceProgress(dlg)
{
   dlg.progress += 1;
   if (dlg.progress > 100) dlg.progress = 0;
   dlg.progressBar.repaint();
   dlg.progressBar.update();
   processEvents();
}


function extractChannels(view)
{
   // retrieve RGBs from view, quiet solution instead of using ChannelExtraction
   var r = r = getNewName(view.id, "_R");
   var g = g = getNewName(view.id, "_G");
   var b = b = getNewName(view.id, "_B");

	for (var i = 0; i < 3; i++)
	{
		view.image.selectedChannel = i;
		var image = new Image(view.image);
		var w = new ImageWindow(image.width, image.height );
		w.mainView.beginProcess( UndoFlag_NoSwapFile );
		w.mainView.image.apply(image);
		w.mainView.endProcess();
		w.mainView.id = [r, g, b][i];
	}

	return [View.viewById(r), View.viewById(g), View.viewById(b)];
}

function copyView( view, newName)
{
   var P = new PixelMath;
   P.expression = "$T";
   P.expression1 = "";
   P.expression2 = "";
   P.expression3 = "";
   P.useSingleExpression = true;
   P.symbols = "";
   P.generateOutput = true;
   P.singleThreaded = false;
   P.use64BitWorkingImage = false;
   P.rescale = false;
   P.rescaleLower = 0;
   P.rescaleUpper = 1;
   P.truncate = true;
   P.truncateLower = 0;
   P.truncateUpper = 1;
   P.createNewImage = true;
   P.showNewImage = false;//true;
   P.newImageId = newName;
   P.newImageWidth = 0;
   P.newImageHeight = 0;
   P.newImageAlpha = false;
   P.newImageColorSpace = PixelMath.prototype.SameAsTarget;
   P.newImageSampleFormat = PixelMath.prototype.SameAsTarget;
   if (P.executeOn(view))
	{
      return View.viewById(newName);
	}
   {
      message("Mask creation failed in function copyView", "Error");
      return null;
   }
}

// ========================================================================
//   Polynomials functions
// ========================================================================


function lstqr()
{
    /* instance variables */
	var pointArray = [];
	var numOfEntries = 0;
	var a = 0;
	var b = 0;
	var c = 0;
	var det = 0;
 	var x = 0;
	var xs = 0;
	var ys = 0;
	/*instance methods */

	this.addPoint = function(x, y)
	{
 	  numOfEntries += 1;
	  pointArray.push([x, y]);
	}

	function aTerm()
	{
	  if (numOfEntries < 3)
	  {
			throw new InvalidOperationException("Insufficient pairs of co-ordinates");
	  }
	  //notation sjk to mean the sum of x_i^j*y_i^k.
	  var s40 = getSx4(); //sum of x^4
	  var s30 = getSx3(); //sum of x^3
	  var s20 = getSx2(); //sum of x^2
	  var s10 = getSx();  //sum of x
	  var s00 = numOfEntries;  //sum of x^0 * y^0  ie 1 * number of entries

	  var s21 = getSx2y(); //sum of x^2*y
	  var s11 = getSxy();  //sum of x*y
	  var s01 = getSy();   //sum of y

	  //a = Da/D
	  return (s21 * (s20 * s00 - s10 * s10) - s11
  				* (s30 * s00 - s10 * s20) + s01
				* (s30 * s10 - s20 * s20))
				/ (s40 * (s20 * s00 - s10 * s10) - s30
				* (s30 * s00 - s10 * s20) + s20
				* (s30 * s10 - s20 * s20));
	}

	/// <summary>
	/// returns the b term of the equation ax^2 + bx + c
	/// </summary>
	/// <returns>b term</returns>
	function bTerm()
	{
	  if (numOfEntries < 3)
	  {
			throw new InvalidOperationException("Insufficient pairs of co-ordinates");
	  }
	  //notation sjk to mean the sum of x_i^j*y_i^k.
	  var s40 = getSx4(); //sum of x^4
	  var s30 = getSx3(); //sum of x^3
	  var s20 = getSx2(); //sum of x^2
	  var s10 = getSx();  //sum of x
	  var s00 = numOfEntries;  //sum of x^0 * y^0  ie 1 * number of entries

	  var s21 = getSx2y(); //sum of x^2*y
	  var s11 = getSxy();  //sum of x*y
	  var s01 = getSy();   //sum of y

	  //b = Db/D
	  return (s40 * (s11 * s00 - s01 * s10) - s30
  				* (s21 * s00 - s01 * s20) + s20
				* (s21 * s10 - s11 * s20))
				/ (s40 * (s20 * s00 - s10 * s10) - s30
				* (s30 * s00 - s10 * s20) + s20
 				* (s30 * s10 - s20 * s20));
	}

	/// <summary>
	/// returns the c term of the equation ax^2 + bx + c
	/// </summary>
	/// <returns>c term</returns>
	function cTerm()
	{
	  if (numOfEntries < 3)
	  {
			throw new InvalidOperationException("Insufficient pairs of co-ordinates");
	  }
	  //notation sjk to mean the sum of x_i^j*y_i^k.
	  var s40 = getSx4(); //sum of x^4
	  var s30 = getSx3(); //sum of x^3
	  var s20 = getSx2(); //sum of x^2
	  var s10 = getSx();  //sum of x
	  var s00 = numOfEntries;  //sum of x^0 * y^0  ie 1 * number of entries

	  var s21 = getSx2y(); //sum of x^2*y
	  var s11 = getSxy();  //sum of x*y
	  var s01 = getSy();   //sum of y

	  //c = Dc/D
	  return (s40 * (s20 * s01 - s10 * s11) - s30
  				* (s30 * s01 - s10 * s21) + s20 * (s30 * s11 - s20 * s21))
				/ (s40 * (s20 * s00 - s10 * s10) - s30
				* (s30 * s00 - s10 * s20) + s20
				* (s30 * s10 - s20 * s20));
	}

	function rSquare() // get r-squared
	{
	  if (numOfEntries < 3)
	  {
			throw new InvalidOperationException("Insufficient pairs of co-ordinates");
	  }
	  // 1 - (total sum of squares / residual sum of squares)
	  return 1 - getSSerr() / getSStot();
	}


	/*helper methods*/
	function getSx() // get sum of x
	{
	  var Sx = 0;
	  for each (var ppair in pointArray)
	  {
			Sx += ppair[0];
	  }
	  return Sx;
	}

	function getSy() // get sum of y
	{
	  var Sy = 0;
	  for each (var ppair in pointArray)
	  {
			Sy += ppair[1];
	  }
	  return Sy;
	}

	function getSx2() // get sum of x^2
	{
	  var Sx2 = 0;
	  for each (var ppair in pointArray)
	  {
			Sx2 += Math.pow(ppair[0], 2); // sum of x^2
	  }
	  return Sx2;
	}

	function getSx3() // get sum of x^3
	{
	  var Sx3 = 0;
	  for each (var ppair in pointArray)
	  {
			Sx3 += Math.pow(ppair[0], 3); // sum of x^3
	  }
	  return Sx3;
	}

	function getSx4() // get sum of x^4
	{
	  var Sx4 = 0;
	  for each (var ppair in pointArray)
	  {
			Sx4 += Math.pow(ppair[0], 4); // sum of x^4
	  }
	  return Sx4;
	}

	function getSxy() // get sum of x*y
	{
	  var Sxy = 0;
	  for each (var ppair in pointArray)
	  {
			Sxy += ppair[0] * ppair[1]; // sum of x*y
	  }
	  return Sxy;
	}

	function getSx2y() // get sum of x^2*y
	{
	  var Sx2y = 0;
	  for each (var ppair in pointArray)
	  {
			Sx2y += Math.pow(ppair[0], 2) * ppair[1]; // sum of x^2*y
	  }
	  return Sx2y;
	}

	function getYMean() // mean value of y
	{
	  var y_tot = 0;
	  for each (var ppair in pointArray)
	  {
			y_tot += ppair[1];
	  }
	  return y_tot / numOfEntries;
	}

	function getSStot() // total sum of squares
	{
	  //the sum of the squares of the differences between
	  //the measured y values and the mean y value
	  var ss_tot = 0;
	  for each (var ppair in pointArray)
	  {
			ss_tot += Math.pow(ppair[1] - getYMean(), 2);
	  }
	  return ss_tot;
	}

	function getSSerr() // residual sum of squares
	{
	  //the sum of the squares of te difference between
	  //the measured y values and the values of y predicted by the equation
	  var ss_err = 0;
	  for each (var ppair in pointArray)
	  {
			ss_err += Math.pow(ppair[1] - getPredictedY(ppair[0]), 2);
	  }
	  return ss_err;
	}

	function getPredictedY(x)
	{
	  //returns value of y predicted by the equation for a given value of x
	  return aTerm() * Math.Pow(x, 2) + bTerm() * x + cTerm();
	}

	this.solve = function()
	{
	  a = aTerm();
	  b = bTerm();
	  c = cTerm();
	  det = b * b - 4 * a * c;
	  if (det<0.0)
	  {
			x = -b / (2 * a);
	  }
	  else if (det > 0.0)
	  {
			var x1 = (-b + Math.sqrt(det)) / (2 * a);
			var x2 = (-b - Math.sqrt(det)) / (2 * a);
			x = (x1 + x2) / 2.0;
	  }
	  else
	  {
			x = (-b + System.Math.sqrt(det)) / (2 * a);
	  }

		xs = -b / ( 2* a);
		ys = (4.0 * a * c - b * b) / (4 * a);
	}

	function Y(x) // get Y(x)
	{
	  return a * x * x + b * x + c;
	}

	this.A = function()
	{
		return a;
	}

	this.B = function()
	{
		return b;
	}

	this.C = function()
	{
		return c;
	}

	this.Det = function()
	{
		return det;
	}

	this.XS = function()
	{
		return xs;
	}

	this.YS = function()
	{
		return ys;
	}
}


function linearViewToBMP(view)
{
	var nlinName = getNewName(view.id, "_preview");
	var nonlinView = copyView(view, nlinName);
	nonlinView.window.hide();
	ApplyAutoSTF ( nonlinView,
						DEFAULT_AUTOSTRETCH_SCLIP,
						DEFAULT_AUTOSTRETCH_TBGND,
						DEFAULT_AUTOSTRETCH_CLINK );
	ApplyHistogram(nonlinView);
	nonlinView.setPropertyValue( Dispose, true );
	var imageBmp = GetWindowBmp(nonlinView.window);
	return imageBmp;
}



function imageMetadata(view)
{
   this.width  = view.image.width;
   this.height = view.image.height;
}

function PreviewControl(parent)
{
   this.__base__ = Frame;
   this.__base__(parent);

   this.SetImage = function(image, metadata)
   {
      this.image = image;
      this.metadata = metadata;
      this.scaledImage = null;
      this.SetZoomOutLimit();
      this.UpdateZoom(-100);
      //this.stf_Button.visible = true;
   }

   this.RefreshImage = function(image)
   {
      this.image = image;
		repaint();
	}

   this.UpdateZoom = function (newZoom, refPoint)
   {
      newZoom = Math.max(this.zoomOutLimit, Math.min(2, newZoom));
      if (newZoom == this.zoom && this.scaledImage)
         return;

      if(refPoint==null)
         refPoint=new Point(this.scrollbox.viewport.width/2, this.scrollbox.viewport.height/2);
      var imgx=null;
      if(this.scrollbox.maxHorizontalScrollPosition>0)
         imgx=(refPoint.x+this.scrollbox.horizontalScrollPosition)/this.scale;
      var imgy=null;
      if(this.scrollbox.maxVerticalScrollPosition>0)
         imgy=(refPoint.y+this.scrollbox.verticalScrollPosition)/this.scale;

      this.zoom = newZoom;
      this.scaledImage = null;
      gc(true);
      if (this.zoom > 0)
      {
         this.scale = this.zoom;
         this.zoomVal_Label.text = format("%d:1", this.zoom);
      }
      else
      {
         this.scale = 1 / (-this.zoom + 2);
         this.zoomVal_Label.text = format("1:%d", -this.zoom + 2);
      }
      if (this.image)
         this.scaledImage = this.image.scaled(this.scale);
      else
         this.scaledImage = {width:this.metadata.width * this.scale, height:this.metadata.height * this.scale};
      this.scrollbox.maxHorizontalScrollPosition = Math.max(0, this.scaledImage.width - this.scrollbox.viewport.width);
      this.scrollbox.maxVerticalScrollPosition = Math.max(0, this.scaledImage.height - this.scrollbox.viewport.height);

      if(this.scrollbox.maxHorizontalScrollPosition>0 && imgx!=null)
         this.scrollbox.horizontalScrollPosition = (imgx*this.scale)-refPoint.x;
      if(this.scrollbox.maxVerticalScrollPosition>0 && imgy!=null)
         this.scrollbox.verticalScrollPosition = (imgy*this.scale)-refPoint.y;

      this.scrollbox.viewport.update();
   }

	this.zoomIn_Button = new ToolButton( this );
	with (this.zoomIn_Button)
	{
   	icon = scaledResource( ":/icons/zoom-in.png" );
   	setScaledFixedSize( 20, 20 );
   	toolTip = "Zoom in";
   	onMousePress = function()
		{
			this.parent.UpdateZoom(this.parent.zoom+1);
		};
	}


   this.zoomOut_Button = new ToolButton( this );
	with (this.zoomOut_Button)
	{
   	icon = scaledResource( ":/icons/zoom-out.png" );
   	setScaledFixedSize( 20, 20 );
   	toolTip = "Zoom in";
   	onMousePress = function()
   	{
      	this.parent.UpdateZoom(this.parent.zoom-1);
   	};
	}


   this.zoom11_Button = new ToolButton( this );
	with (this.zoom11_Button)
	{
   	icon = scaledResource( ":/icons/zoom-1-1.png" );
   	setScaledFixedSize( 20, 20 );
   	toolTip = "Zoom 1:1";
   	onMousePress = function()
		{
			this.parent.UpdateZoom(1);
		};
	}

   this.stf_Button =  new ToolButton( this );
   this.stf_Button.text = "STF";
   this.stf_Button.visible = false;
   this.stf_Button.toolTip = "ScreenTransferFunction";
   this.stf_Button.onMousePress = function()
   {
      var preview = this.parent;
      preview.onCustomSTF.call(this);
		this.parent.repaint();
   }

   this.buttons_Sizer = new HorizontalSizer;
   this.buttons_Sizer.add( this.zoomIn_Button );
   this.buttons_Sizer.add( this.zoomOut_Button );
   this.buttons_Sizer.add( this.zoom11_Button );
   this.buttons_Sizer.addStretch();
   this.buttons_Sizer.add( this.stf_Button );


   this.setScaledMinSize(300,200);
   this.zoom = 1;
   this.scale = 1;
   this.zoomOutLimit = -5;
   this.scrollbox = new ScrollBox(this);
   this.scrollbox.autoScroll = true;
   this.scrollbox.tracking = true;
   this.scrollbox.cursor = new Cursor(StdCursor_Arrow);

   this.scroll_Sizer = new HorizontalSizer;
   this.scroll_Sizer .add( this.scrollbox );

   this.SetZoomOutLimit = function()
   {
      var scaleX = Math.ceil(this.metadata.width/this.scrollbox.viewport.width);
      var scaleY = Math.ceil(this.metadata.height/this.scrollbox.viewport.height);
      var scale = Math.max(scaleX,scaleY);
      this.zoomOutLimit = -scale+2;
   }

   this.scrollbox.onHorizontalScrollPosUpdated = function (newPos)
   {
      this.viewport.update();
   }
   this.scrollbox.onVerticalScrollPosUpdated = function (newPos)
   {
      this.viewport.update();
   }

   this.forceRedraw = function()
   {
      this.scrollbox.viewport.update();
   };

   this.scrollbox.viewport.onMouseWheel = function (x, y, delta, buttonState, modifiers)
   {
      var preview = this.parent.parent;
      preview.UpdateZoom(preview.zoom + (delta > 0 ? -1 : 1), new Point(x,y));
   }

   this.scrollbox.viewport.onMousePress = function ( x, y, button, buttonState, modifiers )
   {
      var preview = this.parent.parent;
      var p =  preview.transform(x, y, preview);
      if(preview.onCustomMouseDown)
      {
         preview.onCustomMouseDown.call(this, p.x, p.y, button, buttonState, modifiers )
      }
   }

   this.scrollbox.viewport.onMouseMove = function ( x, y, buttonState, modifiers )
   {
      var preview = this.parent.parent;
      var p =  preview.transform(x, y, preview);
      preview.Xval_Label.text = p.x.toString();
      preview.Yval_Label.text = p.y.toString();

      if(preview.onCustomMouseMove)
      {
         preview.onCustomMouseMove.call(this, p.x, p.y, buttonState, modifiers )
      }
   }

   this.scrollbox.viewport.onMouseRelease = function (x, y, button, buttonState, modifiers)
   {
      var preview = this.parent.parent;

      var p =  preview.transform(x, y, preview);
      if(preview.onCustomMouseUp)
      {
         preview.onCustomMouseUp.call(this, p.x, p.y, button, buttonState, modifiers )
      }
   }

   this.scrollbox.viewport.onResize = function (wNew, hNew, wOld, hOld)
   {
      var preview = this.parent.parent;
      if(preview.metadata && preview.scaledImage)
      {
         this.parent.maxHorizontalScrollPosition = Math.max(0, preview.scaledImage.width - wNew);
         this.parent.maxVerticalScrollPosition = Math.max(0, preview.scaledImage.height - hNew);
         preview.SetZoomOutLimit();
         preview.UpdateZoom(preview.zoom);
      }
      this.update();
   }

   this.scrollbox.viewport.onPaint = function (x0, y0, x1, y1)
   {
      var preview = this.parent.parent;
      var graphics = new VectorGraphics(this);

      graphics.fillRect(x0,y0, x1, y1, new Brush(0xff202020));
      var offsetX = this.parent.maxHorizontalScrollPosition>0 ? -this.parent.horizontalScrollPosition : (this.width-preview.scaledImage.width)/2;
      var offsetY = this.parent.maxVerticalScrollPosition>0 ? -this.parent.verticalScrollPosition: (this.height-preview.scaledImage.height)/2;
      graphics.translateTransformation(offsetX, offsetY);
      if(preview.image)
         graphics.drawBitmap(0, 0, preview.scaledImage);
      else
         graphics.fillRect(0, 0, preview.scaledImage.width, preview.scaledImage.height, new Brush(0xff000000));

      graphics.pen = new Pen(0xffffffff,0);
      graphics.drawRect(-1, -1, preview.scaledImage.width + 1, preview.scaledImage.height + 1);

      if(preview.onCustomPaint)
      {
         graphics.antialiasing = true;
         graphics.scaleTransformation(preview.scale,preview.scale);
         preview.onCustomPaint.call(this, graphics, x0, y0, x1, y1);
      }
      graphics.end();
   }

   this.transform = function(x, y, preview)
   {
      var scrollbox = preview.scrollbox;
      var ox = 0;
      var oy = 0;
      ox = scrollbox.maxHorizontalScrollPosition>0 ? -scrollbox.horizontalScrollPosition : (scrollbox.viewport.width-preview.scaledImage.width)/2;
      oy = scrollbox.maxVerticalScrollPosition>0 ? -scrollbox.verticalScrollPosition: (scrollbox.viewport.height-preview.scaledImage.height)/2;
      var coordPx = new Point((x - ox) / preview.scale, (y - oy) / preview.scale);
      return new Point(coordPx.x, coordPx.y);
   }

   this.center = function()
   {
      var preview = this;
      var scrollbox = preview.scrollbox;
      var x = scrollbox.viewport.width / 2;
      var y = scrollbox.viewport.height / 2;
      var p =  this.transform(x, y, preview);
      return p;
   }

   this.zoomLabel_Label =new Label(this);
   this.zoomLabel_Label.text = "Zoom:";
   this.zoomVal_Label =new Label(this);
   this.zoomVal_Label.text = "1:1";

   this.Xlabel_Label = new Label(this);
   this.Xlabel_Label .text = "X:";
   this.Xval_Label = new Label(this);
   this.Xval_Label.text = "---";
   this.Ylabel_Label = new Label(this);
   this.Ylabel_Label.text = "Y:";
   this.Yval_Label = new Label(this);
   this.Yval_Label.text = "---";

   this.coords_Frame = new Frame(this);
   this.coords_Frame.backgroundColor = 0xffffffff;
   this.coords_Frame.sizer = new HorizontalSizer;
   this.coords_Frame.sizer.margin = 2;
   this.coords_Frame.sizer.spacing = 4;
   this.coords_Frame.sizer.add(this.zoomLabel_Label);
   this.coords_Frame.sizer.add(this.zoomVal_Label);
   this.coords_Frame.sizer.addSpacing(6);
   this.coords_Frame.sizer.add(this.Xlabel_Label);
   this.coords_Frame.sizer.add(this.Xval_Label);
   this.coords_Frame.sizer.addSpacing(6);
   this.coords_Frame.sizer.add(this.Ylabel_Label);
   this.coords_Frame.sizer.add(this.Yval_Label);

   this.coords_Frame.sizer.addStretch();

   this.sizer = new VerticalSizer;
   this.sizer.add(this.buttons_Sizer);
   this.sizer.add(this.scroll_Sizer);
   this.sizer.add(this.coords_Frame);
}

function GetWindowBmp(window)
{
   var imageOrg = window.mainView.image;
   var tmpW = null;
   try
   {
      tmpW = new ImageWindow(imageOrg.width, imageOrg.height, imageOrg.numberOfChannels,
         window.bitsPerSample, window.isFloatSample, imageOrg.isColor, "Aux");
      tmpW.mainView.beginProcess(UndoFlag_NoSwapFile);
      tmpW.mainView.image.apply(imageOrg);
     // ApplySTF(tmpW.mainView, window.mainView.stf);
      tmpW.mainView.endProcess();
      var bmp = new Bitmap(imageOrg.width, imageOrg.height);
      bmp.assign(tmpW.mainView.image.render());
      return bmp;
   } finally
   {
      tmpW.forceClose();
   }
}

PreviewControl.prototype = new Frame;

// ========================================================================
//      STF Auto Stretch routine
// ========================================================================
//
function ApplyAutoSTF( view, shadowsClipping, targetBackground, rgbLinked )
{
   var stf = new ScreenTransferFunction;

   var n = view.image.isColor ? 3 : 1;

   var median = view.computeOrFetchProperty( "Median" );

   var mad = view.computeOrFetchProperty( "MAD" );
   mad.mul( 1.4826 ); // coherent with a normal distribution

   if ( rgbLinked )
   {
      /*
       * Try to find how many channels look as channels of an inverted image.
       * We know a channel has been inverted because the main histogram peak is
       * located over the right-hand half of the histogram. Seems simplistic
       * but this is consistent with astronomical images.
       */
      var invertedChannels = 0;
      for ( var c = 0; c < n; ++c )
         if ( median.at( c ) > 0.5 )
            ++invertedChannels;

      if ( invertedChannels < n )
      {
         /*
          * Noninverted image
          */
         var c0 = 0, m = 0;
         for ( var c = 0; c < n; ++c )
         {
            if ( 1 + mad.at( c ) != 1 )
               c0 += median.at( c ) + shadowsClipping * mad.at( c );
            m  += median.at( c );
         }
         c0 = Math.range( c0/n, 0.0, 1.0 );
         m = Math.mtf( targetBackground, m/n - c0 );

         stf.STF = [ // c0, c1, m, r0, r1
                     [c0, 1, m, 0, 1],
                     [c0, 1, m, 0, 1],
                     [c0, 1, m, 0, 1],
                     [0, 1, 0.5, 0, 1] ];
      }
      else
      {
         /*
          * Inverted image
          */
         var c1 = 0, m = 0;
         for ( var c = 0; c < n; ++c )
         {
            m  += median.at( c );
            if ( 1 + mad.at( c ) != 1 )
               c1 += median.at( c ) - shadowsClipping * mad.at( c );
            else
               c1 += 1;
         }
         c1 = Math.range( c1/n, 0.0, 1.0 );
         m = Math.mtf( c1 - m/n, targetBackground );

         stf.STF = [ // c0, c1, m, r0, r1
                     [0, c1, m, 0, 1],
                     [0, c1, m, 0, 1],
                     [0, c1, m, 0, 1],
                     [0, 1, 0.5, 0, 1] ];
      }
   }
   else
   {
      /*
       * Unlinked RGB channnels: Compute automatic stretch functions for
       * individual RGB channels separately.
       */
      var A = [ // c0, c1, m, r0, r1
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1] ];

      for ( var c = 0; c < n; ++c )
      {
         if ( median.at( c ) < 0.5 )
         {
            /*
             * Noninverted channel
             */
            var c0 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) + shadowsClipping * mad.at( c ), 0.0, 1.0 ) : 0.0;
            var m  = Math.mtf( targetBackground, median.at( c ) - c0 );
            A[c] = [c0, 1, m, 0, 1];
         }
         else
         {
            /*
             * Inverted channel
             */
            var c1 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) - shadowsClipping * mad.at( c ), 0.0, 1.0 ) : 1.0;
            var m  = Math.mtf( c1 - median.at( c ), targetBackground );
            A[c] = [0, c1, m, 0, 1];
         }
      }

      stf.STF = A;
   }

	var t = false;
	if (t)
	{
		console.writeln( "<end><cbr/><br/><b>", view.fullId, "</b>:" );
		for ( var c = 0; c < n; ++c )
		{
			console.writeln( "channel #", c );
			console.writeln( format( "c0 = %.6f", stf.STF[c][0] ) );
			console.writeln( format( "m  = %.6f", stf.STF[c][2] ) );
			console.writeln( format( "c1 = %.6f", stf.STF[c][1] ) );
		}
	}

   stf.executeOn( view );

   if (t) console.writeln( "<end><cbr/><br/>" );
}

function ApplyHistogram(view)
{

	var stf = view.stf;

	var H = [[  0, 0.0, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0],
				[  0, 0.5, 1.0, 0, 1.0]];;

	if (view.image.isColor)
	{
		for (var c = 0; c < 3; c++)
		{
			H[c][0] = stf[c][1];
			H[c][1] = stf[c][0];
		}
	}
	else
	{
		H[3][0] = stf[0][1];
		H[3][1] = stf[0][0];
	}

	var STF = new ScreenTransferFunction;

	view.stf =  [ // c0, c1, m, r0, r1
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
	[0.00000, 1.00000, 0.50000, 0.00000, 1.00000]
	];

	STF.executeOn(view)

	var HT = new HistogramTransformation;
	HT.H = H;
	HT.executeOn(view);
}

// Helper functions

function getNewName(name, suffix)
{
   var newName = name + suffix;
   let n = 1;
   while (!ImageWindow.windowById(newName).isNull)
   {
      ++n;
      newName = name + suffix + n;
   }
   return newName;
}

function stringOf(chr, count)
{
   var s = new String();
   for (var i = 0; i < count; i++) s = s.concat(chr);
   return s;
}

function message(txt)
{
   var bx = new MessageBox(txt, "Error");
   bx.execute();
}

function getAllMainViews()
{
   var mainViews = [];
   var images = ImageWindow.windows;
   for ( var i in images )
   {
      if (images[i].mainView.isMainView) mainViews.push(images[i].mainView);
   }
   return mainViews;
}

function setDisposal(view, b)
{
	//
	// set/reset disposal flag in view
	//
	view.setPropertyValue(Dispose, b);
}

function houseKeeping()
{
	var views = getAllMainViews();
	for (var i = 0; i < views.length; i++)
	{
		if (views[i].hasProperty( Dispose ))
		{
			var dispose = views[i].propertyValue(Dispose);
			if (dispose)
			{
				// Console.writeln(views[i].id + " closed");
				views[i].window.forceClose();
			}
		}
	}
}


showDialog.prototype = new Dialog;

//////////////////////////////////////////////////////////////////////////////
//
// Main script entry point
//
function main()
{
   // Get access to the current active image window.
	houseKeeping();
   var window = ImageWindow.activeWindow;
   if ( window.isNull )
   {
      message( "No active image" );
   }
   else
   {
      var img = window.currentView.image;
      if (!img.isColor)
      {
         message( "This script for RGB images only" );
      }
      else
      {
         Console.hide();
			Console.show();
		   Console.abortEnabled = true;
         var dialog = new showDialog(window.currentView);
         dialog.execute();
			houseKeeping();
      }
   }
}

main();
